import math
from typing import Callable

import torch
import torch.nn as nn
import torch.nn.functional as F


class MLP(nn.Module):
    """
    Class implementing a standard fully-connected network.
    """

    def __init__(
        self,
        features_in: int,
        features_out: int,
        layers: int = 3,
        units: int = 32,
        activation: Callable[[], nn.Module] = nn.ReLU,
        layer_constructor: Callable[[int, int], nn.Module] = nn.Linear,
    ):
        """
        Args:
            features_in: number of input features
            features_out: number of output features
            layers: number of layers
            units: number of hidden nodes
            activation: function that builds a nn.Module used as activation function
            layer_construction: function used to construct the network layers, given the number of
                input and output features
        """
        super().__init__()
        input_dim = features_in
        layer_list = []
        for i in range(layers - 1):
            layer_list.append(layer_constructor(input_dim, units))
            layer_list.append(activation())
            input_dim = units
        layer_list.append(layer_constructor(input_dim, features_out))
        nn.init.zeros_(layer_list[-1].weight)
        nn.init.zeros_(layer_list[-1].bias)
        self.layers = nn.Sequential(*layer_list)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluates the network.

        Args:
            x: network input, shape (n, features_in)
        Returns:
            network output, shape (n, features_out)
        """
        return self.layers(x)


class StackedMLP(nn.Module):
    """
    Builds multiple independent MLPs that can be efficiently evaluated in parallel.
    """

    def __init__(
        self,
        features_in: int,
        features_out: int,
        channels: int,
        layers: int,
        units: int,
        activation: Callable[[], nn.Module] = nn.ReLU,
        layer_constructor: Callable[[int, int], nn.Module] = nn.Linear,
    ):
        """
        Args:
            features_in: number of input features
            features_out: number of output features
            channels: number of channels
            layers: number of layers
            units: number of hidden nodes
            activation: function that builds a nn.Module used as activation function
            layer_construction: function used to construct the network layers, given the number of
                input and output features
        """
        super().__init__()
        self.channels = channels
        self.activation = activation()
        self.features_out = features_out

        input_dim = features_in
        layer_dims = []
        for i in range(layers - 1):
            layer_dims.append((input_dim, units))
            input_dim = units
        layer_dims.append((input_dim, features_out))

        self.weights = nn.ParameterList(
            [torch.empty((channels, n_out, n_in)) for n_in, n_out in layer_dims]
        )
        self.biases = nn.ParameterList(
            [torch.empty((channels, n_out)) for n_in, n_out in layer_dims]
        )
        self.reset_parameters()

    def reset_parameters(self):
        """
        Initializes the network parameters. The parameters of the last layer are initialized to
        zero. Kaiming uniform initializiation is used for the other layers.
        """
        for ws, bs in zip(self.weights[:-1], self.biases[:-1]):
            for w, b in zip(ws, bs):
                nn.init.kaiming_uniform_(w, a=math.sqrt(5))
                fan_in, _ = nn.init._calculate_fan_in_and_fan_out(w)
                bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                nn.init.uniform_(b, -bound, bound)
        nn.init.zeros_(self.weights[-1])
        nn.init.zeros_(self.biases[-1])

    def _forward_single(self, x: torch.Tensor, channel: int) -> torch.Tensor:
        """
        Evaluates the network for a single channel.

        Args:
            x: network input, shape (n, features_in)
            channel: channel index
        Returns:
            network output, shape (n, features_out)
        """
        if x.shape[0] == 0:
            return x.new_zeros((x.shape[0], self.features_out))
        for w, b in zip(self.weights[:-1], self.biases[:-1]):
            x = self.activation(F.linear(x, w[channel], b[channel]))
        return F.linear(x, self.weights[-1][channel], self.biases[-1][channel])

    def _forward_uniform(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluates the network for equal numbers of samples in every channel.

        Args:
            x: network input, shape (n, features_in)
        Returns:
            network output, shape (n, features_out)
        """
        batch_size = x.shape[0]
        x = x.reshape(self.channels, batch_size // self.channels, x.shape[1])
        for w, b in zip(self.weights[:-1], self.biases[:-1]):
            x = self.activation(torch.baddbmm(b[:, None, :], x, w.transpose(1, 2)))
        return torch.baddbmm(b[:, None, :], x, w.transpose(1, 2)).reshape(
            batch_size, -1
        )

    def forward(
        self,
        x: torch.Tensor,
        channel: list[int] | int | None = None,
    ) -> torch.Tensor:
        """
        Evaluates the network.

        Args:
            x: network input, shape (n, features_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - `list`: list of integers, specifying the number of samples in each channel;
                - `int`: integer specifying a single channel containing all the samples;
                - `None`: all channels contain the same number of samples.
        Returns:
            network output, shape (n, features_out)
        """
        if isinstance(channel, list):
            return torch.cat(
                [
                    self._forward_single(xi, i)
                    for i, xi in enumerate(x.split(channel, dim=0))
                ],
                dim=0,
            )
        elif isinstance(channel, int):
            return self._forward_single(x, channel)
        else:
            return self._forward_uniform(x)



import itertools
import signal
import warnings
from collections.abc import Iterable
from dataclasses import astuple, dataclass
from typing import Any, Callable, Literal

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler

from ..nn import MLP, Distribution, Flow
from .buffer import Buffer
from .integrand import Integrand
from .losses import MultiChannelLoss, kl_divergence, stratified_variance
from .metrics import (
    IntegrationMetrics,
    UnweightingMetrics,
    integration_metrics,
    unweighting_metrics,
)


@dataclass
class TrainingStatus:
    """
    Contains the MadNIS training status to pass it to a callback function.

    Args:
        step: optimization step
        loss: loss from the optimization step
        buffered: whether the optimization was performed on buffered samples
        learning_rate: current learning rate if learning rate scheduler is present
        dropped_channels: number of channels dropped after this optimization step
    """

    step: int
    loss: float
    buffered: bool
    learning_rate: float | None
    dropped_channels: int


@dataclass
class SampleBatch:
    """
    Contains a batch of samples

    Args:
        x: samples generated by the flow, shape (n, dim)
        y: remapped samples returned by the integrand, shape (n, remapped_dim)
        q_sample: probabilities of the samples, shape (n, )
        func_vals: integrand value, shape (n, )
        channels: channels indices for multi-channel integration, shape (n, ), otherwise None
        alphas_prior: prior channel weights, shape (n, channels), or None for single-channel
            integration
        alpha_channel_indices: channel indices if not all prior channel weights are stored,
            otherwise None
        weights: integration weight, shape (n, ). Only set when returned from Integrator.sample
            function, otherwise None.
        alphas: channel weights including learned correction, shape (n, channels). Only set when
            returned from Integrator.sample function, otherwise None.
        zero_counts: channel-wise counts of samples with zero-weights that are not included in the
            batch, shape (channels, ). This field is ignored by most methods, as it behaves
            does not have the batch size as its first dimension
    """

    x: torch.Tensor
    y: torch.Tensor | None
    q_sample: torch.Tensor
    func_vals: torch.Tensor
    channels: torch.Tensor | None
    alphas_prior: torch.Tensor | None = None
    alpha_channel_indices: torch.Tensor | None = None
    weights: torch.Tensor | None = None
    alphas: torch.Tensor | None = None
    zero_counts: torch.Tensor | None = None

    def __iter__(self) -> Iterable[torch.Tensor | None]:
        """
        Returns iterator over the fields of the class
        """
        return iter(astuple(self)[:-1])

    def map(self, func: Callable[[torch.Tensor], torch.Tensor]) -> "SampleBatch":
        """
        Applies function to all fields in the batch that are not None and returns a new SampleBatch

        Args:
            func: function that is applied to all fields in the batch. Expects a tensor as argument
                and returns a new tensor
        Returns:
            Transformed SampleBatch
        """
        return SampleBatch(*(None if field is None else func(field) for field in self))

    def split(self, batch_size: int) -> Iterable["SampleBatch"]:
        """
        Splits up the fields into batches and yields SampleBatch objects for every batch.

        Args:
            batch_size: maximal size of the batches
        Returns:
            Iterator over the batches
        """
        for batch in zip(
            *(
                itertools.repeat(None) if field is None else field.split(batch_size)
                for field in self
            )
        ):
            yield SampleBatch(*batch)

    @staticmethod
    def cat(batches: Iterable["SampleBatch"]) -> "SampleBatch":
        """
        Concatenates multiple batches. If the field zero_counts is not None, the zero_counts of
        all batches are added.

        Args:
            batches: Iterable over SampleBatch objects
        Return:
            New SamplaBatch object containing the concatenated batches
        """
        cat_batch = SampleBatch(
            *(
                None if item[0] is None else torch.cat(item, dim=0)
                for item in zip(*batches)
            )
        )
        if batches[0].zero_counts is not None:
            cat_batch.zero_counts = torch.stack(
                [batch.zero_counts for batch in batches], dim=1
            ).sum(dim=1)
        return cat_batch


class Integrator(nn.Module):
    """
    Implements MadNIS training and integration logic. MadNIS integrators are torch modules, so
    their state can easily be saved and loaded using the torch.save and torch.load methods.
    """

    def __init__(
        self,
        integrand: Callable[[torch.Tensor], torch.Tensor] | Integrand,
        dims: int = 0,
        flow: Distribution | None = None,
        flow_kwargs: dict[str, Any] = {},
        train_channel_weights: bool = True,
        cwnet: nn.Module | None = None,
        cwnet_kwargs: dict[str, Any] = {},
        loss: MultiChannelLoss | None = None,
        optimizer: (
            Optimizer | Callable[[Iterable[nn.Parameter]], Optimizer] | None
        ) = None,
        batch_size: int = 1024,
        learning_rate: float = 1e-3,
        scheduler: LRScheduler | Callable[[Optimizer], LRScheduler] | None = None,
        uniform_channel_ratio: float = 1.0,
        integration_history_length: int = 20,
        drop_zero_integrands: bool = False,
        batch_size_threshold: float = 0.5,
        buffer_capacity: int = 0,
        minimum_buffer_size: int = 50,
        buffered_steps: int = 0,
        max_stored_channel_weights: int | None = None,
        channel_dropping_threshold: float = 0.0,
        channel_dropping_interval: int = 100,
        group_channels_in_loss: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ):
        """
        Args:
            integrand: the function to be integrated. In the case of a simple single-channel
                integration, the integrand function can directly be passed to the integrator.
                In more complicated cases, like multi-channel integrals, use the `Integrand` class.
            dims: dimension of the integration space. Only required if a simple function is given
                as integrand.
            flow: sampling distribution used for the integration. If None, a flow is constructed
                using the `Flow` class. Otherwise, it has to be compatible with a normalizing flow,
                i.e. have the interface defined in the `Distribution` class.
            flow_kwargs: If flow is None, these keyword arguments are passed to the `Flow`
                constructor.
            train_channel_weights: If True, construct a channel weight network and train it. Only
                necessary if cwnet is None.
            cwnet: network used for the trainable channel weights. If None and
                train_channel_weights is True, the cwnet is built using the `MLP` class.
            cwnet_kwargs: If cwnet is None and train_channel_weights is True, these keyword
                arguments are passed to the `MLP` constructor.
            loss: Loss function used for training. If not provided, the KL divergence is chosen in
                the single-channel case and the stratified variance is chosen in the multi-channel
                case.
            optimizer: optimizer for the training. Can be an optimizer object or function that is
                called with the model parameters as argument and returns the optimizer. If None, the
                Adam optimizer is used.
            batch_size: Training batch size
            learning_rate: learning rate used for the Adam optimizer
            scheduler: learning rate scheduler for the training. Can be a learning rate scheduler
                object or a function that gets the optimizer as argument and returns the scheduler.
                If None, a constant learning rate is used.
            uniform_channel_ratio: part of samples in each batch that will be distributed equally
                between all channels, value has to be between 0 and 1.
            integration_history_length: number of batches for which the channel-wise means and
                variances are stored. This is used for stratified sampling during integration, and
                during the training if uniform_channel_ratio is different from one.
            drop_zero_integrands: If True, points with integrand zero are dropped and not used for
                the optimization.
            batch_size_threshold: New samples are drawn until the number of samples is at least
                batch_size_threshold * batch_size.
            buffer_capacity: number of samples that are stored for buffered training
            minimum_buffer_size: minimal size of the buffer to run buffered training
            buffered_steps: number of optimization steps on buffered samples after every online
                training step
            max_stored_channel_weights: number of prior channel weights that are buffered for each
                sample. If None, all prior channel weights are saved, otherwise only those for the
                channels with the largest contributions.
            channel_dropping_threshold: all channels which a cumulated contribution to the
                integrand that is smaller than this threshold are dropped
            channel_dropping_interval: number of training steps after which channel dropping
                is performed
            group_channels_in_loss: If True, grouped channels are treated as a single channel in
                the loss function. This may lead to more stable trainings but also prevents the
                optimization of the channel weights of grouped channels relative to each other.
            device: torch device used for training and integration. If None, use default device.
            dtype: torch dtype used for training and integration. If None, use default dtype.
        """
        super().__init__()

        if not isinstance(integrand, Integrand):
            integrand = Integrand(integrand, dims)
        self.integrand = integrand
        self.multichannel = integrand.channel_count is not None

        if flow is None:
            flow = Flow(
                dims_in=integrand.input_dim,
                channels=integrand.unique_channel_count(),
                **flow_kwargs,
            )
        if cwnet is None and train_channel_weights and self.multichannel:
            cwnet = MLP(integrand.remapped_dim, integrand.channel_count, **cwnet_kwargs)
        if cwnet is None:
            parameters = flow.parameters()
        else:
            parameters = itertools.chain(flow.parameters(), cwnet.parameters())
        if optimizer is None:
            self.optimizer = torch.optim.Adam(parameters, learning_rate)
        elif isinstance(optimizer, Optimizer):
            self.optimizer = optimizer
        else:
            self.optimizer = optimizer(parameters)
        if scheduler is None or isinstance(scheduler, LRScheduler):
            self.scheduler = scheduler
        else:
            self.scheduler = scheduler(self.optimizer)

        self.flow = flow
        self.cwnet = cwnet
        self.batch_size = batch_size
        self.uniform_channel_ratio = uniform_channel_ratio
        self.drop_zero_integrands = drop_zero_integrands
        self.batch_size_threshold = batch_size_threshold
        if loss is None:
            self.loss = stratified_variance if self.multichannel else kl_divergence
        else:
            self.loss = loss

        self.minimum_buffer_size = minimum_buffer_size
        self.buffered_steps = buffered_steps
        self.max_stored_channel_weights = (
            None
            if max_stored_channel_weights is None
            or max_stored_channel_weights >= integrand.channel_count
            else max_stored_channel_weights
        )
        if buffer_capacity > 0:
            channel_count = self.max_stored_channel_weights or integrand.channel_count
            buffer_fields = [
                (integrand.input_dim,),
                None if integrand.remapped_dim is None else (integrand.remapped_dim,),
                (),
                (),
                None if integrand.channel_count is None else (),
                None if not integrand.has_channel_weight_prior else (channel_count,),
                None if self.max_stored_channel_weights is None else (channel_count,),
                None,
                None,
            ]
            buffer_dtypes = [
                None,
                None,
                None,
                None,
                torch.int64,
                None,
                torch.int64,
                None,
                None,
            ]
            self.buffer = Buffer(
                buffer_capacity, buffer_fields, persistent=False, dtypes=buffer_dtypes
            )
        else:
            self.buffer = None
        self.channel_dropping_threshold = channel_dropping_threshold
        self.channel_dropping_interval = channel_dropping_interval
        self.group_channels_in_loss = group_channels_in_loss
        hist_shape = (self.integrand.channel_count or 1,)
        self.integration_history = Buffer(
            integration_history_length,
            [hist_shape, hist_shape, hist_shape],
            dtypes=[None, None, torch.int64],
        )
        self.step = 0
        self.step_type_count = 0
        if self.multichannel:
            self.register_buffer(
                "active_channels_mask",
                torch.ones((self.integrand.channel_count,), dtype=torch.bool),
            )
        # Dummy to determine device and dtype
        self.register_buffer("dummy", torch.zeros((1,)))

        if device is not None:
            self.to(device)
        if dtype is not None:
            self.to(dtype)

    def _get_alphas(self, samples: SampleBatch) -> torch.Tensor:
        """
        Runs the channel weight network and returns the normalized channel weights, taking prior
        channel weights and dropped channels into account.

        Args:
            samples: batch of samples
        Returns:
            channel weights, shape (n, channels)
        """
        if self.cwnet is None:
            if samples.alphas_prior is None:
                return samples.x.new_full(
                    (samples.x.shape[0], self.integrand.channel_count),
                    1 / self.integrand.channel_count,
                )
            return self._restore_prior(samples)

        if samples.alphas_prior is None:
            log_alpha_prior = samples.x.new_zeros(
                (samples.x.shape[0], self.integrand.channel_count)
            )
        else:
            log_alpha_prior = self._restore_prior(samples).log()
        log_alpha = log_alpha_prior.clone()
        mask = samples.func_vals != 0
        y = samples.x if samples.y is None else samples.y
        log_alpha[mask] += self.cwnet(y[mask])
        alpha = torch.zeros_like(log_alpha)
        alpha[:, self.active_channels_mask] = F.softmax(
            log_alpha[:, self.active_channels_mask], dim=1
        )
        return alpha

    def _compute_integral(
        self, samples: SampleBatch
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Computes normalized integrand and channel-wise means, variances and counts

        Args:
            samples: batch of samples
        Returns:
            A tuple containing

              - normalized integrand, shape (n, )
              - channel-wise means of the integral, shape (channels, )
              - channel-wise variances of the integral, shape (channels, )
              - channel-wise number of samples, shape (channels, )
        """
        if self.multichannel:
            alphas = torch.gather(
                self._get_alphas(samples), index=samples.channels[:, None], dim=1
            )[:, 0]
            f_true = alphas * samples.func_vals
            f_div_q = f_true.detach() / samples.q_sample
            counts = torch.bincount(
                samples.channels, minlength=self.integrand.channel_count
            )
            if samples.zero_counts is not None:
                counts += samples.zero_counts
            means = torch.bincount(
                samples.channels,
                weights=f_div_q,
                minlength=self.integrand.channel_count,
            ) / counts.clip(min=1)
            variances = (
                torch.bincount(
                    samples.channels,
                    weights=(f_div_q - means[samples.channels]).square(),
                    minlength=self.integrand.channel_count,
                )
                / counts
            )
        else:
            f_div_q = samples.func_vals / samples.q_sample
            f_true = samples.func_vals
            means = f_div_q.mean(dim=0, keepdim=True)
            counts = torch.full((1,), f_div_q.shape[0], device=means.device)
            variances = f_div_q.var(dim=0, keepdim=True)
        return f_true, means, variances, counts

    def _optimization_step(
        self,
        samples: SampleBatch,
    ) -> tuple[float, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Perform one optimization step of the networks for the given samples

        Args:
            samples: batch of samples
        Returns:
            A tuple containing

              - value of the loss
              - channel-wise means of the integral, shape (channels, )
              - channel-wise variances of the integral, shape (channels, )
              - channel-wise number of samples, shape (channels, )
        """
        self.optimizer.zero_grad()
        # TODO: depending on the loss function and for drop_zero_weights=False, we can encounter
        # zero-weight events here and it might be sufficient to evaluate the flow for events with
        # func_val != 0. That might however give wrong results for other loss functions
        if self.multichannel:
            q_test = self.flow.prob(
                samples.x,
                channel=self.integrand.remap_channels(samples.channels),
            )
        else:
            q_test = self.flow.prob(samples.x, channel=samples.channels)
        f_true, means, variances, counts = self._compute_integral(samples)
        channels = (
            self.integrand.remap_channels(samples.channels)
            if self.group_channels_in_loss
            else samples.channels
        )
        loss = self.loss(f_true, q_test, q_sample=samples.q_sample, channels=channels)
        if loss.isnan().item():
            warnings.warn("nan batch: skipping optimization")
        else:
            loss.backward()
            self.optimizer.step()
        return loss.item(), means, variances, counts

    def _restore_prior(self, samples: SampleBatch) -> torch.Tensor:
        """
        Restores the full prior channel weights if only the largest channel weights and their
        indices were saved.

        Args:
            samples: batch of samples
        Returns:
            Tensor of prior channel weights with shape (n, channels)
        """
        if samples.alpha_channel_indices is None:
            return samples.alphas_prior

        alphas_prior_reduced = samples.alphas_prior
        epsilon = torch.finfo(alphas_prior_reduced.dtype).eps

        # strategy 1: distribute difference to 1 evenly among non-stored channels
        # n_rest = self.integrand.channel_count - self.max_stored_channel_weights
        # alphas_prior = torch.clamp(
        #    (1 - alphas_prior_reduced.sum(dim=1, keepdims=True)) / n_rest,
        #    min=epsilon,
        # ).repeat(1, self.integrand.channel_count)
        # alphas_prior.scatter_(1, samples.alpha_channel_indices, alphas_prior_reduced)
        # return alphas_prior

        # strategy 2: set non-stored channel alphas to epsilon, normalize again
        alphas_prior = alphas_prior_reduced.new_full(
            (alphas_prior_reduced.shape[0], self.integrand.channel_count), epsilon
        )
        alphas_prior.scatter_(1, samples.alpha_channel_indices, alphas_prior_reduced)
        return alphas_prior / alphas_prior.sum(dim=1, keepdims=True)

    def _get_channel_contributions(
        self,
        expect_full_history: bool,
        channel_weight_mode: Literal["variance", "mean"],
    ) -> torch.Tensor:
        """
        Uses the list of saved variances or means to compute the contribution of each channel for
        stratified sampling.

        Args:
            expect_full_history: If True, the integration history has to be full, otherwise uniform
                weights are returned.
            channel_weight_mode: specifies whether the channels are weighted by their mean or
                variance. Note that weighting by mean can lead to problems for non-positive functions
        Returns:
            weights for sampling the channels with shape (channels,)
        """
        min_len = self.integration_history.capacity if expect_full_history else 1
        if self.integration_history.size < min_len:
            return torch.ones(
                self.integrand.channel_count,
                device=self.dummy.device,
                dtype=self.dummy.dtype,
            )
        mean_hist, var_hist, count_hist = self.integration_history
        contrib_hist = mean_hist.abs() if channel_weight_mode == "mean" else var_hist
        count_hist = torch.where(
            contrib_hist.isnan(), np.nan, count_hist.to(contrib_hist.dtype)
        )
        hist_weights = count_hist / count_hist.nansum(dim=0)
        return torch.nansum(hist_weights * contrib_hist, dim=0).sqrt()

    def _disable_unused_channels(self) -> int:
        """
        Determines channels with a total relative contribution below
        ``channel_dropping_threshold``, disables them and removes them from the buffer.

        Returns:
            Number of channels that were disabled
        """
        if self.channel_dropping_threshold == 0.0:
            return 0
        if (self.step + 1) % self.channel_dropping_interval != 0:
            return 0

        mean_hist, _, count_hist = self.integration_history
        count_hist = count_hist.to(mean_hist.dtype)
        mean_hist = torch.nan_to_num(mean_hist)
        hist_weights = count_hist / count_hist.sum(dim=0)
        channel_integrals = torch.nansum(hist_weights * mean_hist, dim=0)
        channel_rel_integrals = channel_integrals / channel_integrals.sum()
        cri_sort, cri_argsort = torch.sort(channel_rel_integrals)
        n_irrelevant = torch.count_nonzero(
            cri_sort.cumsum(dim=0) < self.channel_dropping_threshold
        )
        n_disabled = torch.count_nonzero(
            self.active_channels_mask[cri_argsort[:n_irrelevant]]
        )
        self.active_channels_mask[cri_argsort[:n_irrelevant]] = False
        if self.buffer is not None:
            self.buffer.filter(
                lambda batch: self.active_channels_mask[SampleBatch(*batch).channels]
            )
        return n_disabled

    def _store_samples(self, samples: SampleBatch):
        """
        Stores the generated samples and probabilites for reuse during buffered training. If
        ``max_stored_channel_weights`` is set, the largest channel weights are determined and only
        those and their weights are stored.

        Args:
            samples: Object containing a batch of samples
        """
        if self.buffer is None:
            return

        if (
            self.max_stored_channel_weights is not None
            and self.integrand.has_channel_weight_prior
        ):
            # ensure that the alpha for the channel that the sample was generated with
            # is always stored
            alphas_prior_mod = torch.scatter(
                samples.alphas_prior,
                dim=1,
                index=samples.channels[:, None],
                src=torch.tensor(
                    [[2.0]], device=self.dummy.device, dtype=self.dummy.dtype
                ).expand(*samples.alphas_prior.shape),
            )
            largest_alphas, alpha_indices = torch.sort(
                alphas_prior_mod, descending=True, dim=1
            )
            largest_alphas[:, 0] = torch.gather(
                samples.alphas_prior, dim=1, index=samples.channels[:, None]
            )[:, 0]
            samples.alphas_prior = largest_alphas[
                :, : self.max_stored_channel_weights
            ].clone()
            samples.alpha_channel_indices = alpha_indices[
                :, : self.max_stored_channel_weights
            ].clone()

        self.buffer.store(*samples)

    def _get_channels(
        self,
        n: int,
        channel_weights: torch.Tensor,
        uniform_channel_ratio: float,
        return_counts: bool = False,
    ) -> torch.Tensor:
        """
        Create a tensor of channel indices or number of samples per channel in two steps:
        1. Split up n * uniform_channel_ratio equally among all the channels
        2. Sample the rest of the events from the distribution given by channel_weights
           after correcting for the uniformly distributed samples
        This allows stratified sampling by variance weighting while ensuring stable training
        because there are events in every channel.
        Args:
            n: Number of samples as scalar integer tensor
            channel_weights: Weights of the channels (not normalized) with shape (channels,)
            uniform_channel_ratio: Number between 0.0 and 1.0 to determine the ratio of samples
                that will be distributed uniformly first
            return_counts: If True, return number of samples per channels, otherwise the channel
                indices
        Returns:
            If return_counts is True, Tensor with number of samples per channel, shape (channels,).
            Otherwise, Tensor of channel numbers with shape (n,)
        """
        assert channel_weights.shape == (self.integrand.channel_count,)
        n_active_channels = torch.count_nonzero(self.active_channels_mask)
        uniform_per_channel = int(
            np.ceil(n * uniform_channel_ratio / n_active_channels)
        )
        n_per_channel = torch.full(
            (self.integrand.channel_count,),
            uniform_per_channel,
            device=self.dummy.device,
        )
        n_per_channel[~self.active_channels_mask] = 0

        n_weighted = max(n - n_per_channel.sum(), 0)
        if n_weighted > 0:
            normed_weights = (
                channel_weights / channel_weights[self.active_channels_mask].sum()
            )
            normed_weights[~self.active_channels_mask] = 0.0
            probs = torch.clamp(
                normed_weights - uniform_channel_ratio / n_active_channels, min=0
            )
            n_per_channel += torch.ceil(probs * n_weighted / probs.sum()).int()

        remove_chan = 0
        while n_per_channel.sum() > n:
            if n_per_channel[remove_chan] > 0:
                n_per_channel[remove_chan] -= 1
            remove_chan = (remove_chan + 1) % self.integrand.channel_count
        assert n_per_channel.sum() == n

        if return_counts:
            return n_per_channel

        return torch.cat(
            [
                torch.full((npc,), i, device=self.dummy.device)
                for i, npc in enumerate(n_per_channel)
            ]
        )

    def _get_samples(
        self,
        n: int,
        uniform_channel_ratio: float = 0.0,
        train: bool = False,
        channel_weight_mode: Literal["variance", "mean"] = "variance",
    ) -> SampleBatch:
        """
        Draws samples from the flow and evaluates the integrand

        Args:
            n: number of samples
            uniform_channel_ratio: Number between 0.0 and 1.0 to determine the ratio of samples
                that will be distributed uniformly first
            train: If True, the function is used in training mode, i.e. samples where the integrand
                is zero will be removed if drop_zero_integrands is True
            channel_weight_mode: specifies whether the channels are weighted by their mean or
                variance. Note that weighting by mean can lead to problems for non-positive functions
        Returns:
            Object containing a batch of samples
        """
        channels = (
            self._get_channels(
                n,
                self._get_channel_contributions(train, channel_weight_mode),
                uniform_channel_ratio,
            )
            if self.multichannel
            else None
        )
        channels_remapped = self.integrand.remap_channels(channels)

        batches_out = []
        current_batch_size = 0
        while True:
            with torch.no_grad():
                x, prob = self.flow.sample(
                    n,
                    channel=channels_remapped,
                    return_prob=True,
                    device=self.dummy.device,
                    dtype=self.dummy.dtype,
                )

            weight, y, alphas_prior = self.integrand(x, channels)
            batch = SampleBatch(x, y, prob, weight, channels, alphas_prior)

            if not train:
                current_batch_size += batch.x.shape[0]
            elif self.drop_zero_integrands:
                mask = weight != 0.0
                batch = batch.map(lambda t: t[mask])
                if self.multichannel:
                    batch.zero_counts = torch.bincount(
                        channels[~mask], minlength=self.integrand.channel_count
                    )
                else:
                    batch.zero_counts = torch.full(
                        (1,), torch.count_nonzero(~mask), device=x.device
                    )
                current_batch_size += batch.x.shape[0]
            else:
                current_batch_size += weight.count_nonzero()

            # mask = ~(weight.isnan() | x.isnan().any(dim=1)) & mask
            batches_out.append(batch)

            # check this condition at the end such that the sampling runs at least once
            if current_batch_size > self.batch_size_threshold * n:
                break

        return SampleBatch.cat(batches_out)

    def train_step(self) -> TrainingStatus:
        """
        Performs a single training step

        Returns:
            Training status
        """

        if self.step_type_count == 0:
            buffered = False
            samples = self._get_samples(
                self.batch_size, self.uniform_channel_ratio, train=True
            )
            loss, means, variances, counts = self._optimization_step(samples)
            self._store_samples(samples)
            self.integration_history.store(means[None], variances[None], counts[None])

            if self.buffered_steps != 0 and self.buffer.size > self.minimum_buffer_size:
                self.step_type_count += 1
        else:
            buffered = True
            samples = SampleBatch(*self.buffer.sample(self.batch_size))
            loss, _, _, _ = self._optimization_step(samples)
            self.step_type_count = (self.step_type_count + 1) % (
                self.buffered_steps + 1
            )

        dropped_channels = self._disable_unused_channels()
        status = TrainingStatus(
            step=self.step,
            loss=loss,
            buffered=buffered,
            learning_rate=(
                None if self.scheduler is None else self.scheduler.get_last_lr()[0]
            ),
            dropped_channels=dropped_channels,
        )

        if self.scheduler is not None:
            self.scheduler.step()
        self.step += 1
        return status

    def train(
        self,
        steps: int,
        callback: Callable[[TrainingStatus], None] | None = None,
        capture_keyboard_interrupt: bool = False,
    ):
        """
        Performs multiple training steps

        Args:
            steps: number of training steps
            callback: function that is called after each training step with the training status
                as argument
            capture_keyboard_interrupt: If True, a keyboard interrupt does not raise an exception.
                Instead, the current training step is finished and the training is aborted
                afterwards.
        """
        interrupted = False
        if capture_keyboard_interrupt:

            def handler(sig, frame):
                nonlocal interrupted
                interrupted = True

            old_handler = signal.signal(signal.SIGINT, handler)

        try:
            for _ in range(steps):
                status = self.train_step()
                if callback is not None:
                    callback(status)
                if interrupted:
                    break
        finally:
            if capture_keyboard_interrupt:
                signal.signal(signal.SIGINT, old_handler)

    def integrate(self, n: int, batch_size: int = 100000) -> tuple[float, float]:
        """
        Draws samples and computes the integral.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
        Returns:
            tuple with the value of the integral and the MC integration error
        """
        samples = []
        for n_batch in range(0, n, batch_size):
            samples.append(self._get_samples(min(n - n_batch, batch_size)))
        with torch.no_grad():
            _, means, variances, counts = self._compute_integral(
                SampleBatch.cat(samples)
            )
        self.integration_history.store(means[None], variances[None], counts[None])
        integral = means.sum().item()
        error = torch.nansum(variances / counts).sqrt().item()
        return integral, error

    def integration_metrics(
        self, n: int, batch_size: int = 100000
    ) -> IntegrationMetrics:
        """
        Draws samples and computes metrics for the total and channel-wise integration quality.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
        Returns:
            ``IntegrationMetrics`` object, see its documentation for details
        """
        samples = []
        for n_batch in range(0, n, batch_size):
            samples.append(self._get_samples(min(n - n_batch, batch_size)))
        with torch.no_grad():
            _, means, variances, counts = self._compute_integral(
                SampleBatch.cat(samples)
            )
        self.integration_history.store(means[None], variances[None], counts[None])
        return integration_metrics(means, variances, counts)

    def integral(self) -> tuple[float, float]:
        """
        Returns the current estimate of the integral based on previous training iterations and calls
        to the ``integrate`` function.

        Returns:
            tuple with the value of the integral and the MC integration error
        """
        mean_hist, var_hist, count_hist = self.integration_history
        integrals = mean_hist.sum(dim=1)
        variances = torch.nansum(var_hist / count_hist, dim=1)
        weights = torch.nan_to_num(1 / variances)
        weight_sum = weights.sum()
        integral = torch.sum(weights / weight_sum * integrals).item()
        error = torch.sqrt(1 / weight_sum).item()
        return integral, error

    def unweighting_metrics(
        self,
        n: int,
        batch_size: int = 100000,
        channel_weight_mode: Literal["uniform", "mean", "variance"] = "mean",
    ) -> UnweightingMetrics:
        """
        Draws samples and computes metrics for the total and channel-wise integration quality.
        This function is only suitable for functions that are non-negative everywhere.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
            channel_weight_mode: specifies whether the channels are weighted by their mean,
                variance or uniformly.
        Returns:
            ``UnweightingMetrics`` object, see its documentation for details
        """
        samples = self.sample(n, batch_size, channel_weight_mode)
        return unweighting_metrics(
            samples.weights, samples.channels, self.integrand.channel_count
        )

    def sample(
        self,
        n: int,
        batch_size: int = 100000,
        channel_weight_mode: Literal["uniform", "mean", "variance"] = "variance",
    ) -> SampleBatch:
        """
        Draws samples and computes their integration weight

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
            channel_weight_mode: specifies whether the channels are weighted by their mean,
                variance or uniformly. Note that weighting by mean can lead to problems for
                non-positive functions
        Returns:
            ``SampleBatch`` object, see its documentation for details
        """
        if channel_weight_mode == "uniform":
            uniform_channel_ratio = 1.0
            channel_weight_mode = "variance"
        else:
            uniform_channel_ratio = 0.0

        samples = []
        for n_batch in range(0, n, batch_size):
            batch = self._get_samples(
                min(n - n_batch, batch_size),
                uniform_channel_ratio,
                False,
                channel_weight_mode,
            )
            if self.multichannel:
                with torch.no_grad():
                    batch.alphas = self._get_alphas(batch)
                batch.weights = (
                    torch.gather(batch.alphas, index=batch.channels[:, None], dim=1)[
                        :, 0
                    ]
                    * batch.func_vals
                    / batch.q_sample
                )
            else:
                batch.weights = batch.func_vals / batch.q_sample
            samples.append(batch)
        return SampleBatch.cat(samples)



"""Tests for Normal distributions"""

import unittest

import torch
import torchtestcase

from madnis.nn.flow import Flow


# TODO: Add more. Just to have a starter
class FlowTest(torchtestcase.TorchTestCase):
    def test_log_prob(self):
        batch_size = 10
        input_dim = 5
        flow = Flow(input_dim)
        inputs = torch.randn(batch_size, input_dim)
        log_prob = flow.log_prob(inputs)
        self.assertIsInstance(log_prob, torch.Tensor)
        self.assertEqual(log_prob.shape, torch.Size([batch_size]))
        self.assertFalse(torch.isnan(log_prob).any())
        self.assertFalse(torch.isinf(log_prob).any())

    def test_prob(self):
        batch_size = 10
        input_dim = 5
        flow = Flow(input_dim)
        inputs = torch.randn(batch_size, input_dim)
        prob = flow.prob(inputs)
        log_prob = flow.log_prob(inputs)
        self.assertIsInstance(prob, torch.Tensor)
        self.assertEqual(torch.exp(log_prob), prob)

    def test_transform(self):
        self.eps = 1e-8
        batch_size = 100
        input_dim = 10
        flow = Flow(input_dim)
        inputs = torch.rand(batch_size, input_dim)
        forward, fw_jac = flow.transform(inputs)
        inverse, inv_jac = flow.transform(forward, inverse=True)
        forward2, _ = flow.transform(inverse)
        self.assertIsInstance(forward, torch.Tensor)
        self.assertEqual(inverse, inputs)
        self.assertEqual(forward, forward2)
        self.assertEqual(fw_jac, -inv_jac)


if __name__ == "__main__":
    unittest.main()



""" Implementation of multi-channel loss functions """

from functools import wraps
from typing import Callable

import torch


def dtype_epsilon(tensor: torch.Tensor) -> float:
    return torch.finfo(tensor.dtype).eps


SingleChannelLoss = Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]
MultiChannelLoss = Callable[
    [torch.Tensor, torch.Tensor, torch.Tensor | None, torch.Tensor | None], torch.Tensor
]


def multi_channel_loss(loss: SingleChannelLoss) -> MultiChannelLoss:
    """
    Turns a single-channel loss function into a multi-channel loss function by evaluating it for
    each channel separately and then adding them weighted by TODO weighted by what?

    Args:
        loss: single-channel loss function, that expects the integrand value, test probability and
            sampling probability as arguments
    Returns:
        multi-channel loss function, that expects the integrand value, test probability and,
        optionally, sampling probability and channel indices as arguments.
    """

    # TODO: this unfortunately does not yield the correct signature (with the extra channels argument),
    # so it does not show up in the documentation
    @wraps(loss)
    def wrapped_multi(
        f_true: torch.Tensor,
        q_test: torch.Tensor,
        q_sample: torch.Tensor | None = None,
        channels: torch.Tensor | None = None,
    ) -> torch.Tensor:
        if q_sample is None:
            q_sample = q_test
        if channels is None:
            return loss(f_true, q_test, q_sample)

        loss_tot = 0
        for channel in channels.unique():
            mask = channels == channel
            fi, qti, qsi = f_true[mask], q_test[mask], q_sample[mask]
            ni = mask.count_nonzero()
            loss_tot += ni / q_sample.shape[0] * loss(fi, qti, qsi) if ni > 0 else 0.0
        return loss_tot

    return wrapped_multi


def stratified_variance(
    f_true: torch.Tensor,
    q_test: torch.Tensor,
    q_sample: torch.Tensor | None = None,
    channels: torch.Tensor | None = None,
):
    """
    Computes the stratified variance as introduced in [2311.01548] for two given sets of
    probabilities, ``f_true`` and ``q_test``. It uses importance sampling with a sampling
    probability specified by ``q_sample``.

    Args:
        f_true: normalized integrand values
        q_test: estimated function/probability
        q_sample: sampling probability
        channels: channel indices or None in the single-channel case
    Returns:
        computed stratified variance
    """
    if q_sample is None:
        q_sample = q_test
    if channels is None:
        return _variance(f_true, q_test, q_sample)

    stddev_sum = 0
    abs_integral = 0
    for i in channels.unique():
        mask = channels == i
        fi, qti, qsi = f_true[mask], q_test[mask], q_sample[mask]
        stddev_sum += torch.sqrt(_variance(fi, qti, qsi) + dtype_epsilon(f_true))
        abs_integral += torch.mean(fi.detach().abs() / qsi)

    return (stddev_sum / abs_integral) ** 2


def _variance(
    f_true: torch.Tensor,
    q_test: torch.Tensor,
    q_sample: torch.Tensor,
):
    """
    Computes the variance for two given sets of probabilities, ``f_true`` and ``q_test``. It uses
    importance sampling with a sampling probability specified by ``q_sample``.

    Args:
        f_true: normalized integrand values
        q_test: estimated function/probability
        q_sample: sampling probability
    Returns:
        computed variance
    """
    ratio = q_test / q_sample
    mean = torch.mean(f_true / q_sample)
    sq = (f_true / q_test - mean) ** 2
    return (
        torch.mean(sq * ratio)
        if len(f_true) > 0
        else torch.tensor(0.0, device=f_true.device, dtype=f_true.dtype)
    )


@multi_channel_loss
def kl_divergence(
    f_true: torch.Tensor, q_test: torch.Tensor, q_sample: torch.Tensor
) -> torch.Tensor:
    """
    Computes the Kullback-Leibler divergence for two given sets of probabilities, ``f_true`` and
    ``q_test``. It uses importance sampling, i.e. the estimator is divided by an additional factor
    of ``q_sample``.

    Args:
        f_true: normalized integrand values
        q_test: estimated function/probability
        q_sample: sampling probability
        channels: channel indices or None in the single-channel case
    Returns:
        computed KL divergence
    """
    f_true = f_true.detach().abs()
    f_true /= torch.mean(f_true / q_sample)
    log_q = torch.log(q_test)
    log_f = torch.log(f_true + dtype_epsilon(f_true))
    return torch.mean(f_true / q_sample * (log_f - log_q))


@multi_channel_loss
def rkl_divergence(
    f_true: torch.Tensor, q_test: torch.Tensor, q_sample: torch.Tensor
) -> torch.Tensor:
    """
    Computes the reverse Kullback-Leibler divergence for two given sets of probabilities, ``f_true``
    and ``q_test``. It uses importance sampling, i.e. the estimator is divided by an additional
    factor of ``q_sample``.

    Args:
        f_true: normalized integrand values
        q_test: estimated function/probability
        q_sample: sampling probability
        channels: channel indices or None in the single-channel case
    Returns:
        computed KL divergence
    """
    f_true = f_true.detach().abs()
    f_true /= torch.mean(f_true / q_sample)
    ratio = q_test / q_sample
    log_q = torch.log(q_test)
    log_f = torch.log(f_true + dtype_epsilon(f_true))
    return torch.mean(ratio * (log_q - log_f))



import signal
from dataclasses import dataclass
from typing import Callable, Literal

import numpy as np
import torch

from .integrator import Integrator, SampleBatch
from .metrics import (
    IntegrationMetrics,
    UnweightingMetrics,
    integration_metrics,
    unweighting_metrics,
)


@dataclass
class VegasTrainingStatus:
    """
    Contains the VEGAS training status to pass it to a callback function.

    Args:
        step: optimization step
        variance: integration variance in this iteration. Combined for all channels in the
            multi-channel case
    """

    step: int
    variance: float
    # TODO: maybe add more, like channel-wise statistics?


class VegasPreTraining:
    """
    Implements VEGAS pre-training. It wraps around an ``Integrator`` object and uses its integrand,
    sample buffer and integration history. In addition, it also defines the functions sample,
    integrate, integration_metrics and unweighting_metrics to allow for comparisions between VEGAS
    and MadNIS.
    """

    def __init__(
        self,
        integrator: Integrator,
        bins: int = 64,
        damping: float = 0.7,
    ):
        """ """
        import vegas  # vegas module is optional dependency, so import locally

        # TODO: also maybe replace this with our own VEGAS implementation in pytorch?

        self.integrator = integrator
        self.integrand = integrator.integrand
        self.damping = damping

        if self.integrator.multichannel:
            if self.integrand.channel_grouping is None:
                self.grid_channels = [
                    [channel] for channel in range(self.integrand.channel_count)
                ]
            else:
                self.grid_channels = [
                    group.channel_indices
                    for group in self.integrand.channel_grouping.groups
                ]
        else:
            self.grid_channels = [[0]]
        self.grids = [
            vegas.AdaptiveMap(grid=[[0, 1]] * self.integrand.input_dim, ninc=bins)
            for _ in self.grid_channels
        ]
        self.rng = np.random.default_rng()
        self.step = 0

    def train_step(self, samples_per_channel: int) -> VegasTrainingStatus:
        """
        Performs a single VEGAS training iteration

        Args:
            samples_per_channel: number of training samples per channel
        Returns:
            ``VegasTrainingStatus`` object containing metrics of the training progress
        """
        n_channels = self.integrand.channel_count if self.integrator.multichannel else 1
        input_dim = self.integrand.input_dim
        variances = torch.zeros(n_channels)
        counts = torch.zeros(n_channels)
        means = torch.zeros(n_channels)
        # TODO: maybe implement stratified training instead of uniform channels
        for grid, grid_channels in zip(self.grids, self.grid_channels):
            x = np.empty((samples_per_channel, input_dim), float)
            jac = np.empty(samples_per_channel, float)

            r = self.rng.random((samples_per_channel, input_dim))
            grid.map(r, x, jac)
            x_torch = torch.as_tensor(x, dtype=self.integrator.dummy.dtype)
            if self.integrator.multichannel:
                # TODO: no need for random numbers here
                channels = torch.from_numpy(
                    self.rng.choice(grid_channels, (samples_per_channel,))
                )
            else:
                channels = None
            func_vals, y, alphas = self.integrand(x_torch, channels)
            if self.integrator.multichannel:
                alpha = torch.gather(alphas, index=channels[:, None], dim=1)[:, 0]
                f = jac * func_vals.numpy() * alpha.numpy()
            else:
                f = jac * func_vals.numpy()
            if self.integrator.drop_zero_integrands:
                mask_np = f != 0.0
                f = f[mask_np]
                r = r[mask_np]
                jac = jac[mask_np]
                mask_torch = torch.as_tensor(mask_np)
                if self.integrator.multichannel:
                    zero_counts = torch.bincount(
                        channels[~mask_torch], minlength=self.integrand.channel_count
                    )
                    channels = channels[mask_torch]
                    alphas = alphas[mask_torch]
                else:
                    zero_counts = torch.full((1,), torch.count_nonzero(~mask_torch))
                x_torch = x_torch[mask_torch]
                y = y[mask_torch]
                func_vals = func_vals[mask_torch]
            else:
                zero_counts = None
            grid.add_training_data(r, f**2)
            grid.adapt(alpha=self.damping)

            for chan in grid_channels:
                f_chan = torch.from_numpy(f)
                if self.integrator.multichannel:
                    f_chan = f_chan[channels == chan]
                counts_chan = len(f_chan)
                if zero_counts is not None:
                    counts_chan += zero_counts[chan]
                counts[chan] = counts_chan
                means[chan] = f_chan.sum() / counts_chan
                variances[chan] = (f_chan - means[chan]).square().sum() / counts_chan
            self.integrator._store_samples(
                SampleBatch(
                    x_torch,
                    y,
                    torch.from_numpy(1 / jac),
                    func_vals,
                    channels,
                    alphas,
                    zero_counts=zero_counts,
                ).map(lambda t: t.to(self.integrator.dummy.device))
            )
        self.integrator.integration_history.store(
            means[None], variances[None], counts[None]
        )
        status = VegasTrainingStatus(
            step=self.step,
            variance=torch.sqrt(torch.nansum(variances / counts) * counts.sum()).item()
            / means.sum().item(),
        )
        self.step += 1
        return status

    def train(
        self,
        samples_per_channel: list[int],
        callback: Callable[[VegasTrainingStatus], None] | None = None,
        capture_keyboard_interrupt: bool = False,
    ):
        """
        Performs multiple training steps

        Args:
            samples_per_channel: list of the number of samples per channel, with one entry for every
                training iteration
            callback: function that is called after each training step with the training status
                as argument
            capture_keyboard_interrupt: If True, a keyboard interrupt does not raise an exception.
                Instead, the current training step is finished and the training is aborted
                afterwards.
        """
        interrupted = False
        if capture_keyboard_interrupt:

            def handler(sig, frame):
                nonlocal interrupted
                interrupted = True

            old_handler = signal.signal(signal.SIGINT, handler)

        try:
            for sample_count in samples_per_channel:
                status = self.train_step(sample_count)
                if callback is not None:
                    callback(status)
                if interrupted:
                    break
        finally:
            if capture_keyboard_interrupt:
                signal.signal(signal.SIGINT, old_handler)

    def initialize_integrator(self):
        """
        Initializes the flows in the integrator object using the trained VEGAS grid
        """
        grids_torch = [
            torch.as_tensor(
                grid.extract_grid(),
                device=self.integrator.dummy.device,
                dtype=self.integrator.dummy.dtype,
            )
            for grid in self.grids
        ]
        grids = (
            torch.stack(grids_torch, dim=0)
            if self.integrator.multichannel
            else grids_torch[0]
        )
        self.integrator.flow.init_with_grid(grids)

    def sample(
        self,
        n: int,
        batch_size: int = 100000,
        channel_weight_mode: Literal["uniform", "mean", "variance"] = "variance",
    ) -> SampleBatch:
        """
        Draws samples and computes their integration weight

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
            channel_weight_mode: specifies whether the channels are weighted by their mean,
                variance or uniformly. Note that weighting by mean can lead to problems for
                non-positive functions
        Returns:
            ``SampleBatch`` object, see its documentation for details
        """
        if channel_weight_mode == "uniform":
            uniform_channel_ratio = 1.0
            channel_weight_mode = "variance"
        else:
            uniform_channel_ratio = 0.0
        channel_weights = self.integrator._get_channel_contributions(
            False, channel_weight_mode
        )
        samples_per_channel = self.integrator._get_channels(
            n, channel_weights, uniform_channel_ratio, return_counts=True
        )

        input_dim = self.integrand.input_dim
        samples = []
        for grid, grid_channels in zip(self.grids, self.grid_channels):
            n_samples = samples_per_channel[grid_channels].sum()
            x = np.empty((n_samples, input_dim), float)
            jac = np.empty(n_samples, float)

            r = self.rng.random((n_samples, input_dim))
            grid.map(r, x, jac)
            x_torch = torch.as_tensor(x, dtype=self.integrator.dummy.dtype)
            if self.integrator.multichannel:
                channels = torch.cat(
                    [
                        torch.full((samples_per_channel[channel_index],), channel_index)
                        for channel_index in grid_channels
                    ]
                )
            else:
                channels = None
            func_vals, y, alphas = self.integrand(x_torch, channels)
            alpha = torch.gather(alphas, index=channels[:, None], dim=1)[:, 0]
            f = jac * func_vals.numpy() * alpha.numpy()

            sample_batch = SampleBatch(
                x_torch,
                y,
                torch.from_numpy(1 / jac),
                func_vals,
                channels,
                alphas,
                None,
                torch.from_numpy(f),
                alphas,
            )
            samples.append(
                sample_batch.map(lambda t: t.to(self.integrator.dummy.device))
            )
        return SampleBatch.cat(samples)

    def _compute_integral(
        self, samples: SampleBatch
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if self.integrator.multichannel:
            counts = torch.bincount(
                samples.channels, minlength=self.integrand.channel_count
            )
            means = torch.bincount(
                samples.channels,
                weights=samples.weights,
                minlength=self.integrand.channel_count,
            ) / counts.clip(min=1)
            variances = (
                torch.bincount(
                    samples.channels,
                    weights=(samples.weights - means[samples.channels]).square(),
                    minlength=self.integrand.channel_count,
                )
                / counts
            )
        else:
            means = samples.weights.mean(dim=0, keepdim=True)
            counts = torch.full((1,), samples.weights.shape[0], device=means.device)
            variances = samples.weights.var(dim=0, keepdim=True)
        return means, variances, counts

    def integrate(self, n: int) -> tuple[float, float]:
        """
        Draws samples and computes the integral.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
        Returns:
            tuple with the value of the integral and the MC integration error
        """
        means, variances, counts = self._compute_integral(self.sample(n))
        self.integrator.integration_history.store(
            means[None], variances[None], counts[None]
        )
        integral = means.sum().item()
        error = torch.nansum(variances / counts).sqrt().item()
        return integral, error

    def integration_metrics(
        self, n: int, batch_size: int = 100000
    ) -> IntegrationMetrics:
        """
        Draws samples and computes metrics for the total and channel-wise integration quality.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
        Returns:
            ``IntegrationMetrics`` object, see its documentation for details
        """
        means, variances, counts = self._compute_integral(self.sample(n))
        self.integrator.integration_history.store(
            means[None], variances[None], counts[None]
        )
        return integration_metrics(means, variances, counts)

    def unweighting_metrics(
        self,
        n: int,
        batch_size: int = 100000,
        channel_weight_mode: Literal["uniform", "mean", "variance"] = "mean",
    ) -> UnweightingMetrics:
        """
        Draws samples and computes metrics for the total and channel-wise integration quality.
        This function is only suitable for functions that are non-negative everywhere.

        Args:
            n: number of samples
            batch_size: batch size used for sampling and calling the integrand
            channel_weight_mode: specifies whether the channels are weighted by their mean,
                variance or uniformly.
        Returns:
            ``UnweightingMetrics`` object, see its documentation for details
        """
        samples = self.sample(n, batch_size, channel_weight_mode)
        return unweighting_metrics(
            samples.weights, samples.channels, self.integrand.channel_count
        )



"""
This module containes functions and classes that implement the different types of neural network
architectures necessary for (multi-channel) neural importance sampling.
"""

from .flow import Distribution, Flow
from .mlp import MLP, StackedMLP
from .splines import rational_quadratic_spline, unconstrained_rational_quadratic_spline

__all__ = [
    "unconstrained_rational_quadratic_spline",
    "rational_quadratic_spline",
    "MLP",
    "StackedMLP",
    "Flow",
    "Distribution",
]



from dataclasses import dataclass


@dataclass
class ChannelGroup:
    """
    A group of channels

    Args:
        group_index: index of the group in the list of groups
        target_index: index of the channel that all other channels in the group are mapped to
        channel_indices: indices of the channels in the group
    """

    group_index: int
    target_index: int
    channel_indices: list[int]


@dataclass
class ChannelData:
    """
    Information about a single channel

    Args:
        channel_index: index of the channel
        target_index: index of the channel that it is mapped to
        group: channel group that the channel belongs to
        remapped: True if the channel is remapped to another channel
        position_in_group: index of the channel within its group
    """

    channel_index: int
    target_index: int
    group: ChannelGroup
    remapped: bool
    position_in_group: int


class ChannelGrouping:
    """
    Class that encodes how channels are grouped together for a multi-channel integrand
    """

    def __init__(self, channel_assignment: list[int | None]):
        """
        Args:
            channel_assignment: list with an entry for each channel. If None, the channel is not
                remapped. Otherwise, the index of the channel to which it is mapped.
        """
        group_dict = {}
        for source_channel, target_channel in enumerate(channel_assignment):
            if target_channel is None:
                group_dict[source_channel] = ChannelGroup(
                    group_index=len(group_dict),
                    target_index=source_channel,
                    channel_indices=[source_channel],
                )

        self.channels: list[ChannelData] = []
        self.groups: list[ChannelGroup] = list(group_dict.values())

        for source_channel, target_channel in enumerate(channel_assignment):
            if target_channel is None:
                self.channels.append(
                    ChannelData(
                        channel_index=source_channel,
                        target_index=source_channel,
                        group=group_dict[source_channel],
                        remapped=False,
                        position_in_group=0,
                    )
                )
            else:
                group = group_dict[target_channel]
                self.channels.append(
                    ChannelData(
                        channel_index=source_channel,
                        target_index=target_channel,
                        group=group,
                        remapped=True,
                        position_in_group=len(group.channel_indices),
                    )
                )
                group.channel_indices.append(source_channel)



import math
from abc import ABC, abstractmethod
from typing import Callable, Literal

import numpy as np
import torch
import torch.nn as nn

from .mlp import MLP, StackedMLP
from .splines import unconstrained_rational_quadratic_spline

L2PI = -0.5 * math.log(2 * math.pi)

Mapping = Callable[[torch.Tensor, bool], tuple[torch.Tensor, torch.Tensor]]


class Distribution(ABC):
    """
    Abstract base class for a (potentially learnable) distribution that can be used for sampling and
    density estimation, like a normalizing flow.
    """

    @abstractmethod
    def sample(
        self,
        n: int,
        channel: torch.Tensor | list[int] | int | None = None,
        return_log_prob: bool = False,
        return_prob: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, ...]:
        """
        Draws samples following the distribution

        Args:
            n: number of samples
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
            return_log_prob: if True, also return the log-probabilities
            return_prob: if True, also return the probabilities
            device: device of the returned tensor. Only required if no condition is given.
            dtype: dtype of the returned tensor. Only required if no condition is given.
        Returns:
            samples with shape (n, dims_in). Depending on the arguments ``return_log_prob``,
            ``return_prob``, this function should also return the log-probabilities with shape (n, ),
            the probabilities with shape (n, ).
        """
        pass

    @abstractmethod
    def log_prob(
        self,
        x: torch.Tensor,
        channel: torch.Tensor | list[int] | int | None = None,
    ) -> torch.Tensor:
        """
        Computes the log-probabilities of the input data.

        Args:
            x: input data, shape (n, dims_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
        Returns:
            log-probabilities with shape (n, )
        """
        pass

    @abstractmethod
    def prob(
        self,
        x: torch.Tensor,
        channel: torch.Tensor | list[int] | int | None = None,
        return_latent: bool = False,
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        """
        Computes the probabilities of the input data.

        Args:
            x: input data, shape (n, dims_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
        Returns:
            probabilities with shape (n, )
        """
        pass


class Flow(nn.Module, Distribution):
    """
    Coupling-block based normalizing flow (1605.08803) using rational quadratic spline
    transformations (1906.04032). Both conditional and non-conditional flows are supported. The
    class also allows to build multi-channel flows, i.e. an efficient implementation of multiple
    independent flows with the same hyperparameters.
    """

    def __init__(
        self,
        dims_in: int,
        dims_c: int = 0,
        uniform_latent: bool = True,
        permutations: Literal["log", "random", "exchange"] = "log",
        condition_masks: torch.Tensor | None = None,
        blocks: int | None = None,
        subnet_constructor: Callable[[int, int], nn.Module] | None = None,
        layers: int = 3,
        units: int = 32,
        activation: Callable[[], nn.Module] = nn.ReLU,
        layer_constructor: Callable[[int, int], nn.Module] = nn.Linear,
        channels: int | None = None,
        mapping: Mapping | list[Mapping] | None = None,
        bins: int = 10,
        spline_bounds: float = 10.0,
        min_bin_width: float = 1e-3,
        min_bin_height: float = 1e-3,
        min_bin_derivative: float = 1e-3,
    ):
        """
        Args:
            dims_in: input dimension
            dims_c: condition dimension
            uniform_latent: If True, encode mapping from [0,1]^d to [0,1]^d and use a uniform
                latent space distribution. If False, encode mapping from R^d to R^d and use
                Gaussian latent space distribution.
            permutations: Defines the strategy to permute the input dimensions between coupling
                blocks. "log": logarithmic decomposition, so that every dimension is conditioned on
                every other dimension at least once. "random": randomly permute dimensions.
                "exchange": condition the first half of the input on the second half, then the
                other way around, repeatedly.
            condition_masks: Overwrites the permutation strategy with a custom conditioning mask
                with shape (blocks, dims_in). Components where the mask is True are used as
                condition, and components where it is False are transformed.
            blocks: number of coupling blocks. Only needed if permutations is "random" or
                "exchange"
            subnet_constructor: function used to construct the flow sub-networks, with the
                number of input features and output features of the subnet as arguments. If None,
                the MLP (single channel) or StackedMLP (multi-channel) classes are used.
            layers: number of subnet layers. Only relevant if subnet_constructor=None.
            units: number of subnet hidden nodes. Only relevant if subnet_constructor=None.
            activation: function that builds a nn.Module used as activation function. Only
                relevant if subnet_constructor=None.
            layer_constructor: function used to construct the subnet layers, given the number of
                input and output features. Only relevant if subnet_constructor=None.
            channels: If None, build single-channel flow. If integer, build multi-channel flow
                with this number of channels.
            mapping: Specifies a single mapping function or a list of mapping functions (one per
                channel) that are applied to the input before it enters the flow (forward
                direction) or after drawing samples using the flow (inverse direction). The
                arguments of the function are the input data and a boolean whether the
                transformation is inverted. It must return the transformed value and the logarithm
                of the Jacobian determinant of the transformation.
            bins: number of RQ spline bins
            spline_bounds: If uniform_latent=False, the splines are defined on the interval
                [-spline_bounds, spline_bounds].
            min_bin_width: minimal width of the spline bins
            min_bin_height: minimal height of the spline bins
            min_bin_derivative: minimal derivative at the spline bin edges
        """
        super().__init__()

        self.dims_in = dims_in
        self.dims_c = dims_c
        self.channels = channels
        self.bins = bins
        self.uniform_latent = uniform_latent
        self.mapping = mapping
        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.min_bin_derivative = min_bin_derivative

        if uniform_latent:
            self.spline_low = 0
            self.spline_high = 1
        else:
            self.spline_low = -spline_bounds
            self.spline_high = spline_bounds

        if subnet_constructor is None:
            if channels is None:
                subnet_constructor = lambda features_in, features_out: MLP(
                    features_in,
                    features_out,
                    layers,
                    units,
                    activation,
                    layer_constructor,
                )
            else:
                subnet_constructor = lambda features_in, features_out: StackedMLP(
                    features_in,
                    features_out,
                    channels,
                    layers,
                    units,
                    activation,
                    layer_constructor,
                )

        if condition_masks is None:
            if permutations == "log":
                n_perms = int(np.ceil(np.log2(dims_in)))
                condition_masks = (
                    torch.tensor(
                        [
                            [int(i) for i in np.binary_repr(i, n_perms)]
                            for i in range(dims_in)
                        ]
                    )
                    .flip(dims=(1,))
                    .bool()
                    .t()
                    .repeat_interleave(2, dim=0)
                )
                condition_masks[1::2, :] ^= True
            elif permutations == "exchange":
                condition_masks = torch.cat(
                    (
                        torch.ones(dims_in // 2, dtype=torch.bool),
                        torch.zeros(dims_in - dims_in // 2, dtype=torch.bool),
                    )
                )[None, :].repeat((blocks, 1))
                condition_masks[1::2, :] ^= True
            elif permutations == "random":
                condition_masks = torch.cat(
                    (
                        torch.ones(dims_in // 2, dtype=torch.bool),
                        torch.zeros(dims_in - dims_in // 2, dtype=torch.bool),
                    )
                )[None, :].repeat((blocks, 1))
                for i in range(blocks):
                    condition_masks[i] = condition_masks[i][torch.randperm(dims_in)]
            else:
                raise ValueError(f"Unknown permutation type {permutations}")
        self.condition_masks = condition_masks

        self.subnets = nn.ModuleList()
        for mask in self.condition_masks:
            dims_cond = torch.count_nonzero(mask)
            self.subnets.append(
                subnet_constructor(
                    dims_cond + dims_c, (dims_in - dims_cond) * (3 * bins + 1)
                )
            )

    def _apply_mappings(
        self, x: torch.Tensor, inverse: bool, channel: list[int] | int | None
    ) -> tuple[torch.Tensor, torch.Tensor | float]:
        """
        Applies the single mapping or channel-wise mappings to the input data

        Args:
            x: input data, shape (n, dims_in)
            inverse: if True, use inverted mapping (i.e. the sampling direction)
            channel: list of number of samples per channel, channel index, or None for no channels
                or equally sized channels
        Returns:
            tuple containing the transformed values with shape (n, dims_in), and log Jacobian
            determinants with shape (n, ) of the mapping
        """
        if self.mapping is None:
            return x, 0.0

        if isinstance(self.mapping, list):
            if isinstance(channel, int):
                mapping = self.mapping[channel]
                return mapping(x, inverse)
            else:
                map_x = []
                map_jac = []
                for xc, mapping in zip(x.split(channel, dim=0), self.mapping):
                    xm, jm = mapping(xc, inverse)
                    map_x.append(xm)
                    map_jac.append(jm)

    def transform(
        self,
        x: torch.Tensor,
        channel: torch.Tensor | list[int] | int | None = None,
        inverse: bool = False,
        c: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Transforms the input data into the latent space or back.

        Args:
            x: input data, shape (n, dims_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
            inverse: if True, use inverted transformation (i.e. the sampling direction)
            c: condition, shape (n, dims_c) or None for an unconditional flow
        Returns:
            tuple containing the transformed values with shape (n, dims_in), and log Jacobian
            determinants with shape (n, ) of the transformation
        """
        if isinstance(channel, torch.Tensor):
            channel_perm = torch.argsort(channel)
            x = x[channel_perm]
            c = None if c is None else c[channel_perm]
            channel = channel.bincount(minlength=self.channels).tolist()
        else:
            x = x.clone()
            channel_perm = None

        if inverse:
            jac = 0.0
        else:
            x, jac = self._apply_mappings(x, True, channel)

        if self.channels is None:
            channel_args = ()
        else:
            channel_args = (channel,)

        batch_size = x.shape[0]
        if inverse:
            blocks = zip(reversed(self.condition_masks), reversed(self.subnets))
        else:
            blocks = zip(self.condition_masks, self.subnets)
        for mask, subnet in blocks:
            inv_mask = ~mask
            x_trafo = x[:, inv_mask]
            x_cond = x[:, mask]
            if c is not None:
                x_cond = torch.cat((x_cond, c), dim=1)
            subnet_out = subnet(x_cond, *channel_args).reshape(
                (batch_size, x_trafo.shape[1], -1)
            )
            x_out, block_jac = unconstrained_rational_quadratic_spline(
                x_trafo,
                subnet_out[:, :, : self.bins],
                subnet_out[:, :, self.bins : 2 * self.bins],
                subnet_out[:, :, 2 * self.bins :],
                inverse,
                self.spline_low,
                self.spline_high,
                self.spline_low,
                self.spline_high,
                self.min_bin_width,
                self.min_bin_height,
                self.min_bin_derivative,
            )
            x[:, inv_mask] = x_out
            jac += block_jac.sum(dim=1)

        if inverse:
            x, map_jac = self._apply_mappings(x, False, channel)
            jac += map_jac

        if channel_perm is not None:
            channel_perm_inv = torch.argsort(channel_perm)
            x = x[channel_perm_inv]
            jac = jac[channel_perm_inv]

        return x, jac

    def _latent_log_prob(self, z: torch.Tensor):
        """
        Computes the log-probability of a vector in latent space.

        Args:
            z: latent space vector, shape (n, dims_in)
        Returns:
            log-probabilities, shape (n, )
        """
        if self.uniform_latent:
            return 0.0
        else:
            return z.shape[1] * L2PI - z.square().sum(dim=1) / 2

    def log_prob(
        self,
        x: torch.Tensor,
        channel: torch.Tensor | list[int] | int | None = None,
        c: torch.Tensor | None = None,
        return_latent: bool = False,
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        """
        Computes the log-probabilities of the input data.

        Args:
            x: input data, shape (n, dims_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
            c: condition, shape (n, dims_c) or None for an unconditional flow
            return_latent: if True, also return the latent space vector
        Returns:
            log-probabilities with shape (n, ). If ``return_latent`` is True, it also returns the
            latent space vector with shape (n, dims_in).
        """
        z, jac = self.transform(x, channel, False, c)
        log_prob = self._latent_log_prob(z) + jac
        if return_latent:
            return log_prob, z
        else:
            return log_prob

    def prob(
        self,
        x: torch.Tensor,
        channel: torch.Tensor | list[int] | int | None = None,
        c: torch.Tensor | None = None,
        return_latent: bool = False,
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        """
        Computes the probabilities of the input data.

        Args:
            x: input data, shape (n, dims_in)
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
            c: condition, shape (n, dims_c) or None for an unconditional flow
            return_latent: if True, also return the latent space vector
        Returns:
            probabilities with shape (n, ). If ``return_latent`` is True, it also returns the
            latent space vector with shape (n, dims_in).
        """
        log_prob, z = self.log_prob(x, channel, c, True)
        if return_latent:
            return log_prob.exp(), z
        else:
            return log_prob.exp()

    def sample(
        self,
        n: int | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
        return_log_prob: bool = False,
        return_prob: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        c: torch.Tensor | None = None,
        return_latent: bool = False,
    ) -> torch.Tensor | tuple[torch.Tensor, ...]:
        """
        Draws samples from the probability distribution encoded by the flow.

        Args:
            n: number of samples. Only required if no condition is given.
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
            return_log_prob: if True, also return the log-probabilities
            return_prob: if True, also return the probabilities
            device: device of the returned tensor. Only required if no condition is given.
            dtype: dtype of the returned tensor. Only required if no condition is given.
            c: condition, shape (n, dims_c) or None for an unconditional flow
            return_latent: if True, also return the latent space vector
        Returns:
            samples with shape (n, dims_in). Depending on the arguments ``return_log_prob``,
            ``return_prob`` and ``return_latent``, this function will also return the log-probabilities
            with shape (n, ), the probabilities with shape (n, ) and the latent space vector with
            shape (n, dims_in).
        """
        if n is None:
            n = len(c)
        if c is None:
            options = {}
            if device is not None:
                options["device"] = device
            if dtype is not None:
                options["dtype"] = dtype
        else:
            options = {"device": c.device, "dtype": c.dtype}

        if self.uniform_latent:
            z = torch.rand((n, self.dims_in), **options)
        else:
            z = torch.randn((n, self.dims_in), **options)
        x, jac = self.transform(z, channel, True, c)
        if return_log_prob or return_prob:
            log_prob_latent = self._latent_log_prob(z)
            log_prob = log_prob_latent - jac

        extra_returns = []
        if return_log_prob:
            extra_returns.append(log_prob)
        if return_prob:
            extra_returns.append(log_prob.exp())
        if return_latent:
            extra_returns.append(z)
        if len(extra_returns) > 0:
            return x, *extra_returns
        else:
            return x

    def init_with_grid(self, grid: torch.Tensor):
        """
        Initializes the flow using a VEGAS grid, i.e. from bins with varying width and equal
        probability. The number of bins of this grid should be larger than the number of RQ spline
        bins. This function then performs the bin reduction algorithm as described in [2311.01548].

        Args:
            grid: edges of the VEGAS grid bins with shape (dims_in, vegas_bins+1) for single-channel
                flows or (channels, dims_in, vegas_bins+1) for multi-channel flows
        """
        # Initialize width, heights and derivatives from VEGAS grid
        w = grid.diff(dim=-1)
        h = grid.new_full(w.shape, 1 / w.shape[-1])
        d = (
            torch.cat((w[..., :1], (w[..., :-1] + w[..., 1:]) / 2, w[..., -1:]), dim=-1)
            * w.shape[-1]
        )

        # Run bin reduction algorithm
        while w.shape[-1] > self.bins:
            mask = (w[..., :-1] + w[..., 1:] < 2 / self.bins) & (
                h[..., :-1] + h[..., 1:] < 2 / self.bins
            )
            mask = mask | ~torch.any(mask, dim=-1, keepdim=True)
            diff = torch.abs(
                w[..., 1:] / h[..., 1:] - w[..., :-1] / h[..., :-1],
            ) + 1e9 * (~mask)
            mindiff = torch.argmin(diff, dim=-1, keepdim=True)
            delete_mask = torch.ones(
                w.shape, dtype=torch.bool, device=w.device
            ).scatter_(-1, mindiff, 0)
            next_shape = (*w.shape[:-1], w.shape[-1] - 1)
            w_col = w.gather(-1, mindiff)
            h_col = h.gather(-1, mindiff)
            w = w[delete_mask].reshape(next_shape)
            w.scatter_add_(-1, mindiff, w_col)
            h = h[delete_mask].reshape(next_shape)
            h.scatter_add_(-1, mindiff, h_col)
            d = torch.cat(
                (d[..., :1], d[..., 1:][delete_mask].reshape(next_shape)), dim=-1
            )

        # Invert softmax and softplus functions applied to subnet outputs
        w_unnorm = torch.log(
            (w - self.min_bin_width) / (1 - self.bins * self.min_bin_width)
        )
        w_unnorm -= torch.mean(w_unnorm, dim=-1, keepdim=True)
        h_unnorm = torch.log(
            (h - self.min_bin_height) / (1 - self.bins * self.min_bin_height)
        )
        h_unnorm -= torch.mean(h_unnorm, dim=-1, keepdim=True)
        d_unnorm = torch.log(
            torch.clip(
                torch.exp(
                    (self.min_bin_derivative + math.log(2)) / d
                    - self.min_bin_derivative
                )
                - 1,
                min=self.min_bin_derivative * 1e-5,
            )
        )
        rqs_grid = torch.cat((w_unnorm, h_unnorm, d_unnorm), dim=-1)

        # Initialize weights and biases of the last layer of the subnets of the first
        # two coupling blocks
        if len(grid.shape) == 3:
            weights_0 = self.subnets[0].weights[-1]
            weights_1 = self.subnets[1].weights[-1]
            biases_0 = self.subnets[0].biases[-1]
            biases_1 = self.subnets[1].biases[-1]
        else:
            weights_0 = self.subnets[0].layers[-1].weight
            weights_1 = self.subnets[1].layers[-1].weight
            biases_0 = self.subnets[0].layers[-1].bias
            biases_1 = self.subnets[1].layers[-1].bias
        nn.init.zeros_(weights_0)
        nn.init.zeros_(weights_1)
        with torch.no_grad():
            biases_0.copy_(
                rqs_grid[..., ~self.condition_masks[0], :].flatten(start_dim=-2)
            )
            biases_1.copy_(
                rqs_grid[..., ~self.condition_masks[1], :].flatten(start_dim=-2)
            )



"""
This module contains functions and classes to train neural importance sampling networks and
evaluate the integration and sampling performance.
"""

from .buffer import Buffer
from .channel_grouping import ChannelData, ChannelGroup, ChannelGrouping
from .integrand import Integrand
from .integrator import Integrator, SampleBatch, TrainingStatus
from .losses import (
    kl_divergence,
    multi_channel_loss,
    rkl_divergence,
    stratified_variance,
)
from .metrics import (
    IntegrationMetrics,
    UnweightingMetrics,
    integration_metrics,
    unweighting_metrics,
)
from .vegas_pretraining import VegasPreTraining

__all__ = [
    "Integrator",
    "TrainingStatus",
    "SampleBatch",
    "Integrand",
    "Buffer",
    "multi_channel_loss",
    "stratified_variance",
    "kl_divergence",
    "rkl_divergence",
    "ChannelGroup",
    "ChannelData",
    "ChannelGrouping",
    "UnweightingMetrics",
    "unweighting_metrics",
    "IntegrationMetrics",
    "integration_metrics",
    "VegasPreTraining",
]



from typing import Callable

import torch
import torch.nn as nn

from .channel_grouping import ChannelGrouping


class Integrand(nn.Module):
    """
    Class that wraps an integrand function and meta-data necessary to use advanced MadNIS
    features like learnable multi-channel weights, grouped channels and channel weight priors.
    """

    def __init__(
        self,
        function: Callable,
        input_dim: int,
        bounds: list[list[float]] | None = None,
        channel_count: int | None = None,
        remapped_dim: int | None = None,
        has_channel_weight_prior: bool = False,
        channel_grouping: ChannelGrouping | None = None,
    ):
        """
        Args:
            function: integrand function.
                The signature depends on the other arguments:

                  - single-channel integration, ``channel_count=None``: ``x -> f``
                  - basic multi-channel integration, ``remapped_dim=None``,
                    ``has_channel_weight_prior=False``: ``(x, c) -> f``
                  - with channel weights, ``remapped_dim=None``, ``has_channel_weight_prior=True``:
                    ``(x, c) -> (f, alpha)`` (no trainable channel weights possible)
                  - with channel-dependent mapping, ``remapped_dim: int``,
                    ``has_channel_weight_prior=False``: ``(x, c) -> (f, y)``
                  - all features, ``remapped_dim: int``, ``has_channel_weight_prior=True``:
                    ``(x, c) -> (f, y, alpha)``

                with the following tensors:

                  - ``x`` is a point generated by the importance sampling, shape (n, input_dim),
                  - ``c`` is the channel index, shape (n, ),
                  - ``f`` is the integrand value, shape (n, ),
                  - ``y`` is the point after applying a channel-dependent mapping, shape
                    (n, remapped_dim)
                  - ``alpha`` is the prior channel weight, shape (n, channel_count).
            input_dim: dimension of the integration space
            bounds: List of pairs ``[lower bound, upper bound]`` of the integration interval for
                all dimensions. The integrand is rescaled so that the MadNIS training can be
                performed on the unit hypercube. If None, the unit hypercube is used as integration
                domain.
            channel_count: None in the single-channel case, specifies the number of channels
                otherwise.
            remapped_dim: If different from None, it gives the dimension of a remapped space,
                with a channel-dependent mapping computed as part of the integrand function.
            has_channel_weight_prior: If True, the integrand returns channel weights
            channel_grouping: ChannelGrouping object or None if all channels are independent
        """
        super().__init__()
        self.input_dim = input_dim
        self.remapped_dim = input_dim if remapped_dim is None else remapped_dim
        self.channel_count = channel_count
        self.has_channel_weight_prior = has_channel_weight_prior
        self.channel_grouping = channel_grouping

        if channel_count is None:
            self.function = lambda x, channels: (function(x), None, None)
        elif remapped_dim is None:
            if has_channel_weight_prior:

                def func(x, channels):
                    w, prior = function(x, channels)
                    return w, None, prior

                self.function = func
            else:
                self.function = lambda x, channels: (function(x, channels), None, None)

        elif has_channel_weight_prior:
            self.function = function
        else:

            def func(x, channels):
                w, y = function(x, channels)
                return w, y, None

            self.function = func

        if bounds is not None:
            bounds = torch.tensor(bounds)
            self.register_buffer("scale", bounds[:, 1] - bounds[:, 0])
            self.register_buffer("offset", bounds[:, 0])
            self.register_buffer("scale_det", self.scale.prod())
            old_func = self.function

            def rescaled_func(x, channels):
                w, y, prior = old_func(self.scale * x + self.offset, channels)
                return self.scale_det * w, y, prior

            self.function = rescaled_func

        self.register_buffer(
            "channel_id_map",
            (
                None
                if self.channel_grouping is None
                else torch.tensor(
                    [
                        channel.group.group_index
                        for channel in self.channel_grouping.channels
                    ]
                )
            ),
        )

    def unique_channel_count(self) -> int:
        """
        Returns the number of channels, or, if some channels are grouped together, the number of
        channel groups
        """
        if self.channel_grouping is None:
            return self.channel_count
        else:
            return len(self.channel_grouping.groups)

    def remap_channels(self, channels: torch.Tensor) -> torch.Tensor:
        """
        Remaps channel indices to the indices of their respective channel groups if a
        ``ChannelGrouping`` object was provided, otherwise returns the indices unchanged.

        Args:
            channels: channel indices, shape (n, )
        Returns:
            remapped channel indices, shape (n, )
        """
        if self.channel_grouping is None:
            return channels
        else:
            return self.channel_id_map[channels]

    def forward(
        self, x: torch.Tensor, channels: torch.Tensor | None
    ) -> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None]:
        return self.function(x, channels)



from typing import Callable

import torch
import torch.nn as nn


class Buffer(nn.Module):
    """
    Circular buffer for multiple tensors with different shapes. The class is a torch.nn.Module to
    allow for simple storage.
    """

    def __init__(
        self,
        capacity: int,
        shapes: list[tuple[int, ...]],
        persistent: bool = True,
        dtypes: list[torch.dtype | None] | None = None,
    ):
        """
        Args:
            capacity: maximum number of samples stored in the buffer
            shapes: shapes of the tensors to be stored, without batch dimension. If a shape is
                None, no tensor is stored at that position. This allows for simpler handling of
                optional stored fields.
            persistent: if True, the content of the buffer is part of the module's state_dict
            dtypes: if different from None, specifies the tensors which have a non-standard dtype
        """
        super().__init__()
        self.keys = []
        if dtypes is None:
            dtypes = [None] * len(shapes)
        for i, (shape, dtype) in enumerate(zip(shapes, dtypes)):
            key = f"buffer{i}"
            self.register_buffer(
                key,
                None if shape is None else torch.zeros((capacity, *shape), dtype=dtype),
                persistent,
            )
            self.keys.append(key)
        self.capacity = capacity
        self.size = 0
        self.store_index = 0

    def _batch_slices(self, batch_size: int) -> slice:
        """
        Returns slices that split up the buffer into batches of at most ``batch_size``, respecting
        the buffer size and periodic boundary.
        """
        start = self.store_index
        while start < self.size:
            stop = min(start + batch_size, self.size)
            yield slice(start, stop)
            start = stop
        start = 0
        while start < self.store_index:
            stop = min(start + batch_size, self.store_index)
            yield slice(start, stop)
            start = stop

    def _buffer_fields(self) -> torch.Tensor | None:
        """
        Iterates over the buffered tensors, without removing the padding if the buffer is not full.

        Returns:
            The buffered tensors, or None if a tensor was initialized with shape None
        """
        for key in self.keys:
            yield getattr(self, key)

    def __iter__(self) -> torch.Tensor | None:
        """
        Iterates over the buffered tensors

        Returns:
            The buffered tensors, or None if a tensor was initialized with shape None
        """
        for key in self.keys:
            buffer = getattr(self, key)
            yield None if buffer is None else buffer[: self.size]

    def store(self, *tensors: torch.Tensor | None):
        """
        Adds the given tensors to the buffer. If the buffer is full, the oldest stored samples are
        overwritten.

        Args:
            tensors: samples to be stored. The shapes of the tensors after the batch dimension must
                match the shapes given during initialization. The argument can be None if the
                corresponding shape was None during initialization.
        """
        store_slice1 = None
        for buffer, data in zip(self._buffer_fields(), tensors):
            if data is None:
                continue
            if store_slice1 is None:
                size = min(data.shape[0], self.capacity)
                end_index = self.store_index + size
                if end_index < self.capacity:
                    store_slice1 = slice(self.store_index, end_index)
                    store_slice2 = slice(0, 0)
                    load_slice1 = slice(0, size)
                    load_slice2 = slice(0, 0)
                else:
                    store_slice1 = slice(self.store_index, self.capacity)
                    store_slice2 = slice(0, end_index - self.capacity)
                    load_slice1 = slice(0, self.capacity - self.store_index)
                    load_slice2 = slice(self.capacity - self.store_index, size)
                self.store_index = end_index % self.capacity
                self.size = min(self.size + size, self.capacity)
            buffer[store_slice1] = data[load_slice1]
            buffer[store_slice2] = data[load_slice2]

    def filter(
        self,
        predicate: Callable[[tuple[torch.Tensor | None, ...]], torch.Tensor],
        batch_size: int = 100000,
    ):
        """
        Removes samples from the buffer that do not fulfill the criterion given by the predicate
        function.

        Args:
            predicate: function that returns a mask for a batch of samples, given a tuple with
                all the buffered fields as argument
            batch_size: maximal batch size to limit memory usage
        """
        masks = []
        masked_size = 0
        for batch_slice in self._batch_slices(batch_size):
            mask = predicate(
                tuple(
                    None if t is None else t[batch_slice] for t in self._buffer_fields()
                )
            )
            masked_size += torch.count_nonzero(mask)
            masks.append(mask)
        for buffer in self._buffer_fields():
            if buffer is None:
                continue
            buffer[:masked_size] = torch.cat(
                [
                    buffer[batch_slice][mask]
                    for batch_slice, mask in zip(self._batch_slices(batch_size), masks)
                ],
                dim=0,
            )
        self.size = masked_size
        self.store_index = masked_size % self.capacity

    def sample(self, count: int) -> list[torch.Tensor | None]:
        """
        Returns a batch of samples drawn from the buffer without replacement.

        Args:
            count: number of samples
        Returns:
            samples drawn from the buffer
        """
        weights = next(b for b in self._buffer_fields() if b is not None).new_ones(
            self.size
        )
        indices = torch.multinomial(weights, min(count, self.size), replacement=False)
        return [
            None if buffer is None else buffer[indices]
            for buffer in self._buffer_fields()
        ]



import math
from dataclasses import astuple, dataclass

import torch


@dataclass
class UnweightingMetrics:
    """
    Metrics for the unweighting performance

    Args:
        cut_eff: cut efficiency
        uweff_before_cuts: unweighting efficiency before cuts (computed as
            ``uweff_after_cuts * cut_eff``)
        uweff_before_cuts_partial: unweighting efficiency without over-weights before cuts (computed
            as ``uweff_after_cuts_partial * cut_eff``)
        uweff_after_cuts: unweighting efficiency after cuts
        uweff_after_cuts_partial: unweighting efficiency without over-weights after cuts
        over_weight_rate: fraction of over-weight samples
    """

    cut_eff: float
    uweff_before_cuts: float
    uweff_before_cuts_partial: float
    uweff_after_cuts: float
    uweff_after_cuts_partial: float
    over_weight_rate: float


def unweighting_metrics(
    weights: torch.Tensor,
    channels: torch.Tensor | None = None,
    channel_count: int | None = None,
    replica_count: int = 1000,
) -> UnweightingMetrics | tuple[UnweightingMetrics, list[UnweightingMetrics]]:
    """
    Calculate the unweighting efficiency as discussed in arXiv:2001.10028

    Args:
        weights: weights of the samples
        channels: channel indices of the samples
        channel_count: number of channels
        replica_count: number of replicas, called m in the reference
    Returns:
        An UnweightingMetrics object. In the multi-channel case, it also returns a list of
        UnweightingMetrics objects for all channels.
    """

    if channels is None:
        channel_count = 1
    uweffs = []
    integrals = []
    for channel in range(channel_count):
        w = weights if channel_count == 1 else weights[channels == channel]
        n_total = len(w)
        mask = w != 0.0
        w = w[mask]
        n_accepted = len(w)

        if n_accepted > 0:
            cut_efficiency = n_accepted / n_total
            integrals.append(w.mean().item() * cut_efficiency)
            sample = w[torch.randint(n_accepted, (replica_count, n_total))]
            s_max = sample.amax(dim=1)
            s_mean = sample.mean(dim=1)
            s_max_median = s_max.median()

            s_acc = torch.mean(s_mean / s_max_median).item()
            s_acc_partial = (sample / s_max_median).clip(max=1).mean().item()
            over_weight_rate = (sample > s_max_median).float().mean().item()
        else:
            integrals.append(0.0)
            s_acc = 0.0
            s_acc_partial = 0.0
            over_weight_rate = 0.0
            cut_efficiency = 0.0

        uweffs.append(
            UnweightingMetrics(
                cut_eff=cut_efficiency,
                uweff_before_cuts=s_acc * cut_efficiency,
                uweff_before_cuts_partial=s_acc_partial * cut_efficiency,
                uweff_after_cuts=s_acc,
                uweff_after_cuts_partial=s_acc_partial,
                over_weight_rate=over_weight_rate,
            )
        )

    integral_total = sum(integrals)
    if channel_count > 1:
        return (
            UnweightingMetrics(
                *(
                    sum(
                        integral / integral_total * metric
                        for integral, metric in zip(integrals, channel_metrics)
                    )
                    for channel_metrics in zip(*(astuple(uweff) for uweff in uweffs))
                )
            ),
            uweffs,
        )
    else:
        return uweffs[0]


@dataclass
class IntegrationMetrics:
    """
    Metrics for the integration performance

    Args:
        integral: total integration results
        count: number of integration samples
        error: Monte Carlo integration error
        rel_error: relative integration error
        rel_stddev: relative standard deviation (does not scale with number of samples)
        rel_stddev_opt: optimal relative standard deviation that would have been possible with
            stratified sampling
        channel_integrals: channel-wise integrals
        channel_counts: channel-wise number of samples
        channel_errors: channel-wise integration errors
        channel_rel_errors: channel-wise relative integration errors
        channel_rel_stddevs: channel-wise relative standard deviations
    """

    integral: float
    count: int
    error: float
    rel_error: float
    rel_stddev: float
    rel_stddev_opt: float
    channel_integrals: list[float]
    channel_counts: list[int]
    channel_errors: list[float]
    channel_rel_errors: list[float]
    channel_rel_stddevs: list[float]


def integration_metrics(
    channel_means: torch.Tensor,
    channel_variances: torch.Tensor,
    channel_counts: torch.Tensor,
) -> IntegrationMetrics:
    """
    Calculate metrics for the integration performance

    Args:
        channel_means: channel-wise integrals
        channel_variances: channel-wise variances
        channel_counts: channel-wise sample counts
    Returns:
        An IntegrationMetrics object
    """
    channel_square_errors = channel_variances / channel_counts
    integral = channel_means.sum().item()
    integral_abs = abs(integral)
    count = channel_counts.sum().item()
    error = channel_square_errors.nansum().sqrt().item()
    channel_stddevs = channel_variances.nan_to_num().sqrt()
    channel_errors = channel_square_errors.nan_to_num().sqrt()
    channel_means_abs = channel_means.abs()
    return IntegrationMetrics(
        integral=integral,
        count=count,
        error=error,
        rel_error=error / integral_abs,
        rel_stddev=error * math.sqrt(count) / integral_abs,
        rel_stddev_opt=(channel_stddevs.sum() * channel_stddevs).sum().sqrt().item()
        / integral,
        channel_counts=channel_counts.tolist(),
        channel_integrals=channel_means.tolist(),
        channel_errors=channel_errors.tolist(),
        channel_rel_errors=(channel_errors / channel_means_abs).tolist(),
        channel_rel_stddevs=(channel_stddevs / channel_means_abs).tolist(),
    )
